package ac.essex.ooechs.facedetection.experiments.cbcl;

import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;
import ac.essex.ooechs.facedetection.twostage.TwoStageEvolutionDRS;
import ac.essex.ooechs.facedetection.singlestage.OneStageDetectionProblem;
import ac.essex.ooechs.facedetection.evolved.ga.results.evolved_face.*;
import ac.essex.ooechs.imaging.commons.util.FileIO;
import ac.essex.ooechs.imaging.commons.fast.HaarlikeFeatures;
import ac.essex.ooechs.imaging.commons.HaarRegions;
import ac.essex.gp.Evolve;
import ac.essex.gp.multiclass.BetterDRS;
import ac.essex.gp.multiclass.PCM;
import ac.essex.gp.problems.DataStack;
import ac.essex.gp.individuals.Individual;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.interfaces.console.ConsoleListener;

import java.io.File;
import java.io.IOException;
import java.util.Vector;

/**
 * Tests the difference between two approaches to learning the CBCL data set.
 * The first is using a one stage approach. The second uses pre-defined features
 * found using a GA.
 * <p/>
 * Both experiments are replicated 10 times with different random seeds, and
 * each GP run is 150 generations.
 * <p/>
 * Preliminary experiments show that DRS is the best technique.
 *
 * @author Olly Oechsle, University of Essex, Date: 20-Jun-2008
 * @version 1.0
 */
public class Experiment1_CBCL {

    public static void main(String[] args) throws Exception {

        AbstractDetectionProblem.generations = 150;

        DataSet training = new DataSet("/home/ooechs/Data/faces/train/face/", "/home/ooechs/Data/faces/train/non-face");
        DataSet testing = new DataSet("/home/ooechs/Data/faces/test/face/", "/home/ooechs/Data/faces/test/non-face");

        File directory = new File("/home/ooechs/Desktop/cbcl/");
        directory.mkdirs();                         

        int[] seeds = new int[15];
        for (int i = 0; i < seeds.length; i++) {
            seeds[i] = 2357 + i;
        }

        Experiment1_CBCL exp = new Experiment1_CBCL();

        exp.train(directory, seeds, training);

        //exp.test(directory, seeds, training);

        System.out.println("Training:");
        exp.test2(directory, seeds, training);
        System.out.println("Testing:");
        exp.test2(directory, seeds, testing);

/*        System.out.println("Training:");
        exp.test(directory, seeds, training);
        System.out.println("Testing:");
        exp.test(directory, seeds, testing);*/
        
/*        System.out.println("Training:");
        exp.test3(directory, seeds, training);
        System.out.println("Testing:");
        exp.test3(directory, seeds, testing);*/

    }

    public Vector<Individual> savedOneStageIndividuals = new Vector<Individual>();
    public Vector<Individual> savedTwoStageIndividuals = new Vector<Individual>();

    public void train(File saveDirectory, int[] seeds, DataSet training) {

        System.out.println("ONE STAGE EVOLUTION");

        StringBuffer buffer = new StringBuffer();

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed + " --------------------------");
            Evolve.seed = seed;
            OneStageDetectionProblem p = new OneStageDetectionProblem(training, null);
            Evolve e = new Evolve(p, new ConsoleListener(ConsoleListener.LOW_VERBOSITY));
            e.run();
            Individual ind = e.getBestIndividual();
            savedOneStageIndividuals.add(ind);
                        System.out.println(ind.toJava());
            ind.save(new File(saveDirectory, "onestage_" + i + ".solution"));
            buffer.append(p.test(ind));
            try {
                FileIO.saveToFile(buffer.toString(), new File(saveDirectory, "onestage.txt"));
            } catch (Exception err) {
                err.printStackTrace();
            }
            System.gc();
        }

        System.out.println("TWO STAGE EVOLUTION");
        buffer.setLength(0);

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed + " --------------------------");
            Evolve.seed = seed;
            TwoStageEvolutionCBCL p = new TwoStageEvolutionCBCL(training, null);
            Evolve e = new Evolve(p, new ConsoleListener(ConsoleListener.LOW_VERBOSITY));
            e.run();
            Individual ind = e.getBestIndividual();
            savedTwoStageIndividuals.add(ind);
            ind.save(new File(saveDirectory, "twostage_" + i + ".solution"));
            buffer.append(p.test(ind));
            try {
                FileIO.saveToFile(buffer.toString(), new File(saveDirectory, "twostage.txt"));
            } catch (Exception err) {
                err.printStackTrace();
            }
            System.gc();
        }

        System.out.println("Done :)");

    }

    public void test(File saveDirectory, int[] seeds, DataSet testing) throws Exception {

        System.out.println("DISK SAVED INDIVIDUALS");

        System.out.println("ONE STAGE EVOLUTION");

        for (int i = 0; i < seeds.length; i++) {
            OneStageDetectionProblem p = new OneStageDetectionProblem(testing, null);
            Individual ind = Individual.load(new File(saveDirectory, "onestage_" + i + ".solution"));

            System.out.print(p.test(ind));
        }

/*        System.out.println("TWO STAGE EVOLUTION");

        for (int i = 0; i < seeds.length; i++) {
            TwoStageEvolutionCBCL p = new TwoStageEvolutionCBCL(testing, null);
            Individual ind = Individual.load(new File(saveDirectory, "twostage_" + i + ".solution"));
            System.out.print(p.test(ind));
        }*/

    }

    public void test2(File saveDirectory, int[] seeds, DataSet testing) throws Exception {

        System.out.println("MEMORY SAVED INDIVIDUALS");

        System.out.println("ONE STAGE EVOLUTION");

        for (int j = 0; j < savedOneStageIndividuals.size(); j++) {
            Individual ind =  savedOneStageIndividuals.elementAt(j);
            OneStageDetectionProblem p = new OneStageDetectionProblem(testing, null);
            System.out.print(p.test(ind));
        }

        System.out.println("TWO STAGE EVOLUTION");

        for (int j = 0; j < savedTwoStageIndividuals.size(); j++) {
            Individual ind =  savedTwoStageIndividuals.elementAt(j);
            TwoStageEvolutionCBCL p = new TwoStageEvolutionCBCL(testing, null);
            System.out.print(p.test(ind));
        }

        System.out.println("Done :)");

    }

    public void test3(File saveDirectory, int[] seeds, DataSet testing) throws Exception {

        System.out.println("JAVA SAVED INDIVIDUAL");

        System.out.println("ONE STAGE EVOLUTION");

        Individual ind =  new Detector();
        OneStageDetectionProblem p = new OneStageDetectionProblem(testing, null);
        System.out.print(p.test(ind));


    }

    class Detector extends Individual {

        public Detector() {
            super(null,0);
        }

        public double execute(DataStack dataStack) {
            HaarlikeFeatures image = dataStack.getIntegralImage().getHaarlikeFeatures();
            HaarlikeFeatures varianceimage = dataStack.getIntegralImage().getHaarlikeFeaturesVariance();
            return eval(0,0,image, varianceimage);
        }

PCM pcm0 = new BetterDRS(-162.5,171.0,new int[]{0,-1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,-1,1,-1,0,1});;
public double eval(int x, int y, HaarlikeFeatures image, HaarlikeFeatures varianceimage) {
    double node1 = varianceimage.getTwoRectangleFeature(x + 2, y + 7, 1, 1, HaarRegions.HORIZONTALLY_ADJACENT);
    double node6 = image.getThreeRectangleFeature(x + 4, y + 3, 1, 1, HaarRegions.VERTICALLY_ADJACENT);
    return pcm0.getClassFromOutput(node1 - node6);
}

    }


    class TwoStageEvolutionCBCL extends TwoStageEvolutionDRS {

        public TwoStageEvolutionCBCL(DataSet training, DataSet testing) {
            super(training, testing);
        }

        public void registerParams(GPParams params) {
            params.registerNode(new TwoRectFeature1());
            params.registerNode(new TwoRectFeature2());
            params.registerNode(new TwoRectFeature3());
            params.registerNode(new TwoRectFeature4());
            params.registerNode(new TwoRectFeature5());
            params.registerNode(new TwoRectFeature6());
            params.registerNode(new TwoRectFeature7());
            params.registerNode(new TwoRectFeature8());
            params.registerNode(new TwoRectFeature9());
            params.registerNode(new TwoRectFeature10());

            params.registerNode(new ThreeRectFeature1());
            params.registerNode(new ThreeRectFeature2());
            params.registerNode(new ThreeRectFeature3());
            params.registerNode(new ThreeRectFeature4());
            params.registerNode(new ThreeRectFeature5());
            params.registerNode(new ThreeRectFeature6());
            params.registerNode(new ThreeRectFeature7());
            params.registerNode(new ThreeRectFeature8());
            params.registerNode(new ThreeRectFeature9());
            params.registerNode(new ThreeRectFeature10());

            params.registerNode(new FourRectFeature1());
            params.registerNode(new FourRectFeature2());
            params.registerNode(new FourRectFeature3());
            params.registerNode(new FourRectFeature4());
            params.registerNode(new FourRectFeature5());
            params.registerNode(new FourRectFeature6());
            params.registerNode(new FourRectFeature7());
            params.registerNode(new FourRectFeature8());
            params.registerNode(new FourRectFeature9());
            params.registerNode(new FourRectFeature10());

        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature2());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature10());

        params.registerNode(new ThreeRectVarFeature1());
        params.registerNode(new ThreeRectVarFeature2());
        params.registerNode(new ThreeRectVarFeature3());
        params.registerNode(new ThreeRectVarFeature4());
        params.registerNode(new ThreeRectVarFeature5());
        params.registerNode(new ThreeRectVarFeature6());
        params.registerNode(new ThreeRectVarFeature7());
        params.registerNode(new ThreeRectVarFeature8());
        params.registerNode(new ThreeRectVarFeature9());
        params.registerNode(new ThreeRectVarFeature10());

        params.registerNode(new FourRectVarFeature1());
        params.registerNode(new FourRectVarFeature2());
        params.registerNode(new FourRectVarFeature3());
        params.registerNode(new FourRectVarFeature4());
        params.registerNode(new FourRectVarFeature5());
        params.registerNode(new FourRectVarFeature6());
        params.registerNode(new FourRectVarFeature7());
        params.registerNode(new FourRectVarFeature8());
        params.registerNode(new FourRectVarFeature9());
        params.registerNode(new FourRectVarFeature10());

        }

    }

}
