package ac.essex.ooechs.facedetection.singlestage;

import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;
import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.util.IntegralTrainingImage;
import ac.essex.ooechs.facedetection.singlestage.nodes.*;
import ac.essex.gp.Evolve;
import ac.essex.gp.multiclass.PCM;
import ac.essex.gp.multiclass.BetterDRS;
import ac.essex.gp.multiclass.CachedOutput;
import ac.essex.gp.problems.DataStack;
import ac.essex.gp.individuals.Individual;
import ac.essex.gp.nodes.*;
import ac.essex.gp.nodes.ercs.SmallIntERC;
import ac.essex.gp.nodes.ercs.LargeIntERC;
import ac.essex.gp.nodes.ercs.PercentageERC;
import ac.essex.gp.nodes.ercs.CustomRangeParameterERC;
import ac.essex.gp.nodes.math.Pow;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.params.NodeConstraints;
import ac.essex.gp.interfaces.console.ConsoleListener;

import java.util.Vector;

/**
 * <p/>
 * In this experiment we want to see how good GP is at solving
 * a face detection problem using a data set.
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 12-Jun-2008
 * @version 1.0
 */
public class OneStageDetectionProblemBoolean extends AbstractDetectionProblem {

    protected int FPCounts[];
    protected int totalFP = 0;

    double bestFitness = Double.MAX_VALUE;

    static boolean weighting = false;
    static boolean optimisation = false;

    public OneStageDetectionProblemBoolean(DataSet training, DataSet testing) {
        super(training, testing);
    }

    public String getName() {
        return "Face Detection Problem - Fixed Features";
    }

    public void initialise(Evolve evolve, GPParams params) {

        params.registerNode(new IF_FP());

        params.registerNode(new Add());
        params.registerNode(new Sub());
        params.registerNode(new Mul());
        params.registerNode(new Div());

        params.registerNode(new Pow());
        params.registerNode(new Min());
        params.registerNode(new Max());

        params.registerNode(new SmallIntERC());
        params.registerNode(new LargeIntERC());
        params.registerNode(new PercentageERC());

        params.registerNode(new PercentageERC(NodeConstraints.PARAMETER));

/*        params.registerNode(new OneRectangleFeature());
        params.registerNode(new OneRectangleVarianceFeature());*/

        params.registerNode(new TwoHorizontalRectangleFeature());
        params.registerNode(new TwoVerticalRectangleFeature());

        params.registerNode(new ThreeHorizontalRectangleFeature());
        params.registerNode(new ThreeVerticalRectangleFeature());

        params.registerNode(new FourHorizontalRectangleFeature());
        params.registerNode(new FourVerticalRectangleFeature());

        params.registerNode(new TwoHorizontalRectangleVarianceFeature());
        params.registerNode(new TwoVerticalRectangleVarianceFeature());

        params.registerNode(new ThreeHorizontalRectangleVarianceFeature());
        params.registerNode(new ThreeVerticalRectangleVarianceFeature());

        params.registerNode(new FourHorizontalRectangleVarianceFeature());
        params.registerNode(new FourVerticalRectangleVarianceFeature());

        /*    params.registerNode(new EvolvedTwoRectangleFeature());
                params.registerNode(new Rectangle());
                params.registerNode(new FixedValueTerminal(0, NodeConstraints.FEATURE));
        */
        params.setReturnType(NodeConstraints.NUMBER);

    }

    public void customiseParameters(GPParams params) {
        params.setGenerations(generations);
        params.setERCOptimisationEnabled(optimisation);
    }


    public void onGenerationStart() {
        super.onGenerationStart();
        if (FPCounts != null) {
            // redistribute the weights
            for (int i = 0; i < trainingData.size(); i++) {
                IntegralTrainingImage trainingImage = trainingData.elementAt(i);
                // add the one to avoid divide by zero and ensure that everything has some weight
                int mistakeCount = FPCounts[i] + 1;

                // adapt the training image weight towards the new value
                double newValue = (mistakeCount * trainingData.size()) / (double) totalFP;

                double difference = newValue - trainingImage.weight;

                difference *= 0.33;

                // learning rate
                if (weighting) trainingImage.weight += difference;

                //trainingImage.weight = (mistakeCount * trainingData.size()) / (double) totalFP;
                //System.out.println("Training Image Weight: " + trainingImage.weight);
            }
        }
        FPCounts = new int[trainingData.size()];
        totalFP = 0;
    }

    public void evaluate(Individual individual, DataStack dataStack, Evolve evolve) {

        // parameters for calculating fitness
        double totalTrueWeight = 0;
        double totalFalseWeight = 0;
        double trueWeight = 0;
        double falseWeight = 0;

        int totalTrue = 0;
        int totalFalse = 0;
        int TP = 0;
        int FP = 0;

        int hits = 0;
        int mistakes = 0;

        dataStack.usesImaging = false;

        // go through all the training data
        for (int i = 0; i < trainingData.size(); i++) {

            IntegralTrainingImage trainingImage = trainingData.elementAt(i);

            dataStack.setIntegralImage(trainingImage.getImage());

            double result = individual.execute(dataStack);

            boolean returnedTrue = result > 0;

            result = returnedTrue? DataSet.TRUE : DataSet.FALSE;

            if (result == trainingImage.classID) {
                hits++;
            } else {
                mistakes++;
                FPCounts[i]++;
                totalFP++;
            }

            if (trainingImage.classID == 1) {
                totalTrueWeight += trainingImage.weight;
                totalTrue++;
                if (returnedTrue) {
                    trueWeight += trainingImage.weight;
                    TP++;
                }
            } else {
                totalFalseWeight += trainingImage.weight;
                totalFalse++;
                if (returnedTrue) {
                    falseWeight += trainingImage.weight;
                    FP++;
                }
            }

        }

        double x = 1;

        double truePercentage = trueWeight / totalTrueWeight;
        double falsePercentage = falseWeight / totalFalseWeight;
        double weightedFitness = x * (1 - truePercentage) + falsePercentage;
        //double weightedFitness = falseWeight / totalFalseWeight;

   /*     double unweightedTruePercentage = TP / (double) totalTrue;
        double unweightedFalsePercentage = FP / (double) totalFalse;
        double unweightedFitness = x * (1 - unweightedTruePercentage) + unweightedFalsePercentage;
*/
        individual.setKozaFitness(weightedFitness);
        individual.setAlternativeFitness(mistakes);

/*

        if (unweightedFitness < bestFitness) {
            bestFitness = unweightedFitness;
            bestIndividual = individual;
        }
*/


        individual.setHits(hits);
        individual.setMistakes(mistakes);

    }
}
